package de.renew.net;

import de.renew.net.event.TransitionEventListener;
import de.renew.net.event.TransitionEventListenerSet;
import de.renew.net.event.TransitionEventProducer;

import de.renew.util.RenewObjectOutputStream;

import de.renew.net.arc.*;

import java.io.IOException;
import java.io.Serializable;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;


/**
 * I represent a transition with all its semantic information,
 * but without any state. I can have a name, an uplink,
 * and a set of inscriptions. My main job is to store
 * all this information. I do not actually perform any
 * useful services.
 */
public class Transition implements Serializable, TransitionEventProducer {
    static final long serialVersionUID = -7955261691016471176L;

    /**
     * The net element id.
     * @serial
     **/
    NetElementID id;

    /**
     * My name.
     **/
    private final String name;

    /**
     * true, if my instances should output a trace message when they
     * occur. This is the default.
     **/
    private boolean trace;

    /**
     * My uplink or null, if I am spontaneous.
     **/
    UplinkInscription uplink;

    /**
     * A set of inscriptions. I expect all inscription objects to be
     * of the type TransitionInscription.
     *
     * @see de.renew.net.TransitionInscription
     **/
    Set inscriptions;

    // ---- TransitionEvent Handling ----------------------------------------


    /**
     * The listeners to notify if transition events occur.
     *
     * This object is used to synchronize listener additions/removals
     * and notification.
     **/
    private transient TransitionEventListenerSet listeners = new TransitionEventListenerSet();

    /**
     * I (a transition) am created as a spontaneous transition.
     * Later on, inscriptions may be added, even an uplink.
     * I am automatically registered at the net as a
     * net element.
     *
     * @param net the net to which this transition should be added.
     */
    public Transition(Net net, String name) {
        id = new NetElementID();
        this.name = name;


        // Trace actions.
        trace = true;


        // Clear the uplink: no uplink is the default.
        uplink = null;


        // Initialize the array for the inscriptions.
        inscriptions = new HashSet();


        // Notify the net that we are done and want to be
        // inserted into the list of transitions.
        net.add(this);
    }

	Transition(String name){
		this.name=new String(name);
		inscriptions = new HashSet();
	}

    /**
     * Returns the net element id.
     * @return The net element id.
     */
    public NetElementID getID() {
        return id;
    }

    /**
     * I return my name as my string representation.
     * My net instances use my name as a prefix for their name.
     * @return my name.
     **/
    public String toString() {
        return name;
    }

    /**
     * Switch my trace flag on or off.
     *
     * @param trace true, if tracing is desired
     **/
    public void setTrace(boolean trace) {
        this.trace = trace;
    }

    /**
     * Am I being traced?
     *
     * @return true, if my trace flag is set
     **/
    public boolean getTrace() {
        return trace;
    }

    /**
     * Am I spontaneous?
     *
     * @return true, if I am spontaneous, i.e., if I do not
     *   have an uplink and if I am not manual
     **/
    public boolean isSpontaneous() {
        return uplink == null
               && !inscriptions.contains(ManualInscription.getInstance());
    }

    public boolean listensToChannel(String channel) {
        return uplink != null && uplink.name.equals(channel);
    }

    /**
     * Here I extract my uplink, if any, from my inscriptions.
     * If there is more than one uplink, it is undetermined which
     * uplink I will use. This method is called after every
     * update to my set of inscriptions.
     **/
    private void checkUplink() {
        // Maintain the count of uplinks and select an uplink
        // from the set of inscriptions. By allowing multiple inscriptions
        // the embedding into a graphical editor becomes easier.
        // On the other hand, the transition might get into
        // an inconsistent state. We try to make it as simulatable as
        // possible, however.
        uplink = null;
        Iterator iterator = inscriptions.iterator();
        while (iterator.hasNext()) {
            Object inscription = iterator.next();
            if (inscription instanceof UplinkInscription) {
                uplink = (UplinkInscription) inscription;
            }
        }
    }

    /**
     * Give me another inscription. After inserting
     * the inscriptions, I will recheck my uplink.
     **/
    public void add(TransitionInscription transitionInscription) {
        if (!inscriptions.contains(transitionInscription)) {
            inscriptions.add(transitionInscription);
            checkUplink();
        }
    }

    /**
     * Remove one my current inscriptions. Check if an
     * uplink is still provided.
     **/
    public void remove(TransitionInscription transitionInscription) {
        if (inscriptions.contains(transitionInscription)) {
            inscriptions.remove(transitionInscription);
            checkUplink();
        }
    }

	//added by david
	//if the operation is sucessfull, then  return true, or return false;
	public synchronized boolean removeArc(PlaceInstance placeInstance,int arcType){
		try{
		
		Iterator iterator=inscriptions.iterator();
		while(iterator.hasNext()){
			TransitionInscription insrc=(TransitionInscription)iterator.next();
			if(insrc instanceof Arc){	
				Arc arc=(Arc)insrc;	
				if(placeInstance.getPlace()==arc.getPlace()&&arcType==arc.getType()){
					iterator.remove();	
					boolean result2=(arc.getPlace()).removeArc(arc);	
					if(result2==true){
						return true;
					}
				}
			}
		}
		}catch(Exception e){
			System.out.println("Transition.removeArc "+e.toString());
		}
		System.out.println("Transition.removeArc failed");
		return false;

	}
	
	public  TransitionInscription getArc(Place place,int arcType){
		
		Iterator iterator=inscriptions.iterator();
		while(iterator.hasNext()){
			TransitionInscription insrc=(TransitionInscription)iterator.next();
			if(insrc instanceof Arc){	
				Arc arc=(Arc)insrc;	
				if(place==arc.getPlace()&&arcType==arc.getType()){
					return arc;
				}
			}
		}
		return null;
	}
	//added by david end 

    /**
     * Serialization method, behaves like default writeObject
     * method except using the domain trace feature, if the
     * output is a RenewObjectOutputStream.
     * @see de.renew.util.RenewObjectOutputStream
     **/
    private void writeObject(java.io.ObjectOutputStream out)
            throws IOException {
        RenewObjectOutputStream rOut = null;
        if (out instanceof RenewObjectOutputStream) {
            rOut = (RenewObjectOutputStream) out;
        }
        if (rOut != null) {
            rOut.beginDomain(this);
        }
        out.defaultWriteObject();
        if (rOut != null) {
            rOut.endDomain(this);
        }
    }

    /**
     * Deserialization method, behaves like default readObject
     * method except restoring additional transient fields.
     **/
    private void readObject(java.io.ObjectInputStream in)
            throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        listeners = new TransitionEventListenerSet();
    }

    public void addTransitionEventListener(TransitionEventListener listener) {
        listeners.addTransitionEventListener(listener);
    }

    public void removeTransitionEventListener(TransitionEventListener listener) {
        listeners.removeTransitionEventListener(listener);
    }

    TransitionEventListenerSet getListenerSet() {
        return listeners;
    }
	//added by david
  	public String getName(){
	 return name;
  	}

	public Transition clone(){
		Transition targetTransition=new Transition(getName());
		targetTransition.id=new NetElementID(id.getFigureID(),id.getGroupID());
		targetTransition.trace=trace;
		Iterator iterator=inscriptions.iterator();
		while(iterator.hasNext()){
			targetTransition.add((TransitionInscription)(iterator.next()));
		}

		targetTransition.listeners=(TransitionEventListenerSet)(listeners.clone());
		targetTransition.uplink=uplink;

		return targetTransition;
 

	}

	public boolean checkArcs(){
		Iterator iterator=inscriptions.iterator();
		TransitionInscription inscr;
		while(iterator.hasNext()){
			inscr=(TransitionInscription)(iterator.next());
			if(inscr instanceof Arc||inscr instanceof ClearArc||inscr instanceof FlexibleArc)
				return false;
					
		}
		return true;
	}
	public Set getConnectedPlaces(){
		Iterator iterator=inscriptions.iterator();
		HashSet set =new HashSet();
		TransitionInscription inscr;
		while(iterator.hasNext()){
			inscr=(TransitionInscription)(iterator.next());
			if(inscr instanceof Arc)
				set.add(((Arc)inscr).getPlace());
					
		}
		return set;
	}
	//added by david end 
}
